<?php
/*
 * Copyright (C) 2017 PayKeeper
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * @author      PayKeeper
 * @copyright   2017 PayKeeper
 * @license     http://opensource.org/licenses/gpl-2.0.php GNU General Public License, version 2 (GPL-2.0)
 */

namespace PayKeeper\PayKeeper\Model;

/**
 * Checkout Payment Method Model Class
 * Class Checkout
 * @package PayKeeper\PayKeeper\Model\Method
 */
class Payment extends \Magento\Payment\Model\Method\AbstractMethod
{
    const CODE = 'paykeeper_checkout';
    /**
     * Checkout Method Code
     */
    protected $_code = self::CODE;

    protected $_canOrder                    = true;
    protected $_isGateway                   = true;
    protected $_canCapture                  = true;
    protected $_canCapturePartial           = true;
    protected $_canRefund                   = true;
    protected $_canCancelInvoice            = true;
    protected $_canVoid                     = true;
    protected $_canRefundInvoicePartial     = true;
    protected $_canAuthorize                = true;
    protected $_isInitializeNeeded          = false;

    private $fiscal_cart = array(); //fz54 cart
    private $order_total = 0; //order total sum
    private $shipping_price = 0; //shipping price
    private $use_taxes = false;
    private $use_delivery = false;
    private $order_params = NULL;
    private $discounts = array();

    /**
     * Get Instance of the Magento Code Logger
     * @return \Psr\Log\LoggerInterface
     */
    protected function getLogger()
    {
        return $this->_logger;
    }

    /**
     * Checkout constructor.
     * @param \Magento\Framework\Model\Context $context
     * @param \Magento\Framework\App\Action\Context $actionContext
     * @param \Magento\Framework\Registry $registry
     * @param \Magento\Framework\Api\ExtensionAttributesFactory $extensionFactory
     * @param \Magento\Framework\Api\AttributeValueFactory $customAttributeFactory
     * @param \Magento\Payment\Helper\Data $paymentData
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Magento\Payment\Model\Method\Logger $logger
     * @param \Magento\Store\Model\StoreManagerInterface $storeManager
     * @param \Magento\Checkout\Model\Session $checkoutSession
     * @param \PayKeeper\PayKeeper\Helper\Data $moduleHelper
     * @param \Magento\Framework\Model\ResourceModel\AbstractResource|null $resource
     * @param \Magento\Framework\Data\Collection\AbstractDb|null $resourceCollection
     * @param array $data
     */
    public function __construct(
        \Magento\Framework\Model\Context $context,
        \Magento\Framework\App\Action\Context $actionContext,
        \Magento\Framework\Registry $registry,
        \Magento\Framework\Api\ExtensionAttributesFactory $extensionFactory,
        \Magento\Framework\Api\AttributeValueFactory $customAttributeFactory,
        \Magento\Payment\Helper\Data $paymentData,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Magento\Payment\Model\Method\Logger  $logger,
        \Magento\Store\Model\StoreManagerInterface $storeManager,
        \Magento\Checkout\Model\Session $checkoutSession,
        \Magento\Sales\Model\Order $order,
        \Magento\Sales\Model\Order\Email\Sender\OrderSender $orderSender,
        \Magento\Sales\Api\OrderRepositoryInterface $orderRepository,
        \Magento\Framework\Api\SearchCriteriaBuilder $searchCriteriaBuilder,
        \Magento\Sales\Model\Service\InvoiceService $invoiceService,
        \Magento\Sales\Model\Order\Email\Sender\InvoiceSender $invoiceSender,
        \Magento\Framework\DB\Transaction $transaction,
        \Magento\Framework\App\Request\Http $request,
        \Magento\Quote\Api\CartRepositoryInterface $quoteRepository,
        \Magento\Quote\Model\QuoteFactory $quoteFactory,
        \PayKeeper\PayKeeper\Helper\Data $moduleHelper,
        \PayKeeper\PayKeeper\Helper\Payment $paymentHelper,
        \Magento\Framework\Model\ResourceModel\AbstractResource $resource = null,
        \Magento\Framework\Data\Collection\AbstractDb $resourceCollection = null,
        array $data = []
    ) {
        $this->orderRepository = $orderRepository;
        parent::__construct(
            $context,
            $registry,
            $extensionFactory,
            $customAttributeFactory,
            $paymentData,
            $scopeConfig,
            $logger,
            $resource,
            $resourceCollection,
            $data
        );
        $this->_actionContext = $actionContext;
        $this->_storeManager = $storeManager;
        $this->_checkoutSession = $checkoutSession;
        $this->_order = $order;
        $this->orderSender = $orderSender;
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
        $this->transaction = $transaction;
        $this->invoiceService = $invoiceService;
        $this->invoiceSender = $invoiceSender;
        $this->_request = $request;
        $this->_quoteFactory = $quoteFactory;
        $this->_quoteRepository = $quoteRepository;
        $this->_moduleHelper = $moduleHelper;
        $this->_paymentHelper = $paymentHelper;
    }

    private function getOrder($orderid = Null)
    {
        if ($orderid) {
            $searchCriteria = $this->searchCriteriaBuilder->addFilter(
                "increment_id",
                $orderid,
                "eq"
                )->create();
            $orderList = $this->orderRepository->getList($searchCriteria)->getItems();
            return reset($orderList);
            //return $orderList;
            //return $this->orderRepository->get($orderid);
        }
        else return $this->_checkoutSession->getLastRealOrder();
    }

    public function getCallback()
    {
  
        // check request type
        if (!$this->_request->isPost())
            die('Wrong request type (not Post)');
  
        // get request variables
        $request = $this->_request->getPost();
        if (empty($request)) {
            die('Request doesn\'t contain POST elements.');
        }

        $id		    =  $request['id'];
        $sum	    =  $request['sum'];
        $clientid	=  $request['clientid'];
        $orderid	=  $request['orderid'];
        $sign	    =  $request['key'];

        // check order id
        if (empty($orderid) || strlen($orderid) > 50)
            die('Missing or invalid order ID');
  
        // check orderid
//        if ($orderid != $order->getId())
//            die('Order not found');
  
        //check price
//        $our_price = number_format($this->_order->getBaseGrandTotal(),2,'.','');
//        if ($our_price != $sum)
//            die('Incorrect amount');
  
        //$this->_paymentInst = $this->_order->getPayment()->getMethodInstance();
        //$our_key = $this->_paymentInst->getConfigData('security_key');
        $secret_key = $this->_paymentHelper->getConfig("payment/paykeeper_checkout/secret_key");

        $check_sign = md5($id.$sum.$clientid.$orderid.$secret_key);
  
            // check transaction signature
        if($sign != $check_sign)
          die('Message digest incorrect');

        $order = $this->getOrder($orderid);
  
        //$this->_order->getPayment()->setTransactionId($id)->setLastTransId($id);
  
        //if ($this->_order->canInvoice()) {
/*        $invoiceId = Mage::getModel('sales/order_invoice_api')->create($this->_order->getIncrementId(), array());
        $invoice = Mage::getModel('sales/order_invoice')->loadByIncrementId($invoiceId);
        $invoice->capture()->save(); */
        //}


//        $this->_order->setData('state', $this->_paymentInst->getConfigData('order_status'));
//        $this->_order->setData('status', $this->_paymentInst->getConfigData('order_status'));
//        $this->_order->save();
        $invoice = $this->invoiceService->prepareInvoice($order);
        $invoice->setRequestedCaptureCase(\Magento\Sales\Model\Order\Invoice::CAPTURE_ONLINE);
        $invoice->register();
        $invoice->save();
        $transactionSave = $this->transaction->addObject(
            $invoice
            )->addObject(
            $invoice->getOrder()
        );
        $transactionSave->save();
        //$this->invoiceSender->send($invoice);
        //send notification code
        //$order->addStatusHistoryComment(
            //__('Notified customer about invoice #%119.', $invoice->getId())
        //)
        //->setIsCustomerNotified(true)
        //->save();
        $this->orderSender->send($orderid);
        $order->save();
  
//        $invoice = $order->prepareInvoice();
//        $invoice->register()->capture();
//        $this->transactionFactory->addObject($invoice)
//            ->addObject($invoice->getOrder())
//            ->save();
//        $order->save();

//        if ($this->_order->canInvoice() && $this->_order->getTotalPaid() == 0) {
//          $invoice = $this->_order->prepareInvoice();
//          $invoice->register()->capture();
//          Mage::getModel('core/resource_transaction')
//            ->addObject($invoice)
//            ->addObject($invoice->getOrder())
//            ->save();
//          //$this->_order->sendNewOrderEmail();
//          //$this->_order->setEmailSent(true);
//          $this->_order->save();
//        }
  
        //Mage::getResourceModel('sales/order_status_collection')->toOptionHash();
  
        $hash = md5($id.$secret_key);
  
        die("OK $hash");

    }


    public function getFormFields()
    {
        $session = $this->_checkoutSession;
        $quote = $session->getQuote();
        $order = $this->getOrder();

        //set order parameters
        $this->setOrderParams(
                $order->getGrandTotal(),                    //sum
                $order->getCustomerFirstName() . " " .
                $order->getCustomerLastName(),              //clientid
                $order->getIncrementId(),                   //orderid
                $order->getCustomerEmail(),                 //client_email
                $order->getBillingAddress()->getTelephone(),//client_phone
                "",                                         //service_name
                $this->getConfigData("form_url"), 
                $this->getConfigData("secret_key")
        );

        //GENERATE FZ54 CART
        $product_cart_sum = 0;
        $tax = "none";
        $tax_sum = 0;

        $cart_data = $order->getAllItems();

        foreach ($cart_data as $item) {
            $sum = 0;
            $tax_rate = 0;
            $taxes = array("tax" => "none", "tax_sum" => 0);

            $name = $item->getName();
            $qty = (int) $item->getQtyOrdered();
            $price = $item->getPrice()+($item->getTaxAmount()/$qty);
            $sum = number_format($price*$qty, 2, ".", "");
            $tax_rate = $item->getTaxPercent();
            if ($tax_rate > 0) {
                $taxes = $this->setTaxes($sum, $tax_rate);
                $this->setUseTaxes();
            }
            $product_cart_sum += $sum;

            $this->updateFiscalCart($this->getPaymentFormType(), 
                $name, $price, $qty, $sum, $taxes["tax"], $taxes["tax_sum"]);
        }

        //add shipping service
        $shipping_tax_rate = 0;
        $shipping_taxes = array("tax" => "none", "tax_sum" => 0);

        $this->setShippingPrice($order->getShippingAmount());
        if ($this->getShippingPrice() > 0) {
            if ($this->getUseTaxes()){
                $shipping_taxes = $this->setTaxes($this->getShippingPrice(),18);
            }
            $this->setUseDelivery(); //for precision correct check
            $this->updateFiscalCart($this->getPaymentFormType(),
                                    $order->getShippingDescription(), 
                                    $this->getShippingPrice(), 1, 
                                    $this->getShippingPrice(), 
                                    $shipping_taxes["tax"],
                                    $shipping_taxes["tax_sum"]);
        }

        //set discounts
        $discount_amount = $this->getOrder()->getDiscountAmount();
        if ($discount_amount < 0) {
            $cart_sum = $this->setDiscounts($product_cart_sum);
        }

        //handle possible precision problem
        $this->correctPrecision($product_cart_sum);

        $fiscal_cart_encoded = json_encode($this->getFiscalCart());

        $to_hash = $this->getOrderTotal(TRUE)            .
                   $this->getOrderParams("clientid")     .
                   $this->getOrderParams("orderid")      .
                   $this->getOrderParams("service_name") .
                   $this->getOrderParams("client_email") .
                   $this->getOrderParams("client_phone") .
                   $this->getOrderParams("secret_key");
        $sign = hash ('sha256' , $to_hash);

        return array ('form_url' =>         $this->getOrderParams("form_url"),
                      'form_type' =>        $this->getPaymentFormType(),
                      'params' => array(
                          "sum" =>          $this->getOrderTotal(TRUE),
                          "clientid" =>     $this->getOrderParams("clientid"),
                          "orderid" =>      $this->getOrderParams("orderid"),
                          "client_email" => $this->getOrderParams("client_email"),
                          "client_phone" => $this->getOrderParams("client_phone"),
                          "phone" =>        $this->getOrderParams("client_phone"),
                          "service_name" => $this->getOrderParams("service_name"),
                          "cart" =>         $fiscal_cart_encoded,
                          "sign" =>         $sign,
        ));

    }

    private function setOrderParams($order_total = 0, $clientid="", $orderid="", $client_email="", 
                                    $client_phone="", $service_name="", $form_url="", $secret_key="")
    {
       $this->setOrderTotal($order_total);
       $this->order_params = array(
           "sum" => $order_total,
           "clientid" => $clientid,
           "orderid" => $orderid,
           "client_email" => $client_email,
           "client_phone" => $client_phone,
           "phone" => $client_phone,
           "service_name" => $service_name,
           "form_url" => $form_url,
           "secret_key" => $secret_key,
       );
    }

    private function getOrderParams($value)
    {
        return array_key_exists($value, $this->order_params) ? $this->order_params["$value"] : False;
    }

    private function updateFiscalCart($ftype, $name="", $price=0, $quantity=0, $sum=0, $tax="none", $tax_sum=0)
    {
        //update fz54 cart
        if ($ftype === "create") {
            $name = str_replace("\n ", "", $name);
            $name = str_replace("\r ", "", $name);
            $name = preg_replace('/"{1,}/','\'',$name);
            $name = htmlspecialchars($name);
        }
        $this->fiscal_cart[] = array(
            "name" => $name,
            "price" => $price,
            "quantity" => $quantity,
            "sum" => $sum,
            "tax" => $tax,
            "tax_sum" => number_format($tax_sum, 2, ".", "")
        );
    }

    private function getFiscalCart()
    {
        return $this->fiscal_cart;
    }

    private function setDiscounts($cart_sum)
    {
        //set discounts
        $discount_modifier_value = ($this->getOrderTotal() - $this->getShippingPrice())/$cart_sum;
        if ($discount_modifier_value < 1) {
            $fiscal_cart_count = ($this->getShippingPrice() != 0) ?
                                            count($this->getFiscalCart())-1 :
                                            count($this->getFiscalCart());
            //iterate fiscal cart without shipping
            for ($pos=0; $pos<$fiscal_cart_count; $pos++) {
                $this->fiscal_cart[$pos]["sum"] *= $discount_modifier_value;
                $this->fiscal_cart[$pos]["price"] = $this->fiscal_cart[$pos]["sum"]/$this->fiscal_cart[$pos]["quantity"];
                //formatting
                $this->fiscal_cart[$pos]["price"] = number_format($this->fiscal_cart[$pos]["price"], 3, ".", "");
                $this->fiscal_cart[$pos]["sum"] = number_format($this->fiscal_cart[$pos]["sum"], 2, ".", "");
                //recalculate taxes
                $this->recalculateTaxes($pos);
            }
        }
    }

    private function correctPrecision($cart_sum)
    {
        //handle possible precision problem
        $fiscal_cart_sum = $cart_sum;
        $total_sum = $this->getOrderTotal(TRUE);
        //add shipping sum to cart sum
        if ($this->getShippingPrice() > 0)
            $fiscal_cart_sum += $this->fiscal_cart[count($this->fiscal_cart)-1]['sum'];
        //debug_info
        //echo "total: " . $total_sum . " - cart: " . $cart_sum;
        $diff_sum = $fiscal_cart_sum - $total_sum;
        if (abs($diff_sum) <= 0.01) {
            $this->setOrderTotal(number_format($total_sum+$diff_sum, 2, ".", ""));
        }
        else {
            if ($this->getUseDelivery() && ($fiscal_cart_sum < $total_sum)) {
                $this->setOrderTotal(number_format($total_sum+$diff_sum, 2, ".", ""));
                $delivery_pos = count($this->getFiscalCart())-1;
                $this->fiscal_cart[$delivery_pos]["price"] = number_format(
                                   $this->fiscal_cart[$delivery_pos]["price"]+$diff_sum, 2, ".", "");
                $this->fiscal_cart[$delivery_pos]["sum"] = number_format(
                                   $this->fiscal_cart[$delivery_pos]["sum"]+$diff_sum, 2, ".", "");
                $this->recalculateTaxes($delivery_pos);
            }
        }
    }

    private function setOrderTotal($value)
    {
        $this->order_total = $value;
    }

    private function getOrderTotal($format=False)
    {
        return ($format) ? number_format($this->order_total, 2, ".", "") : 
                                         $this->order_total;
    }

    private function setShippingPrice($value)
    {
        $this->shipping_price = $value;
    }

    private function getShippingPrice()
    {
        return $this->shipping_price;
    }

    private function getPaymentFormType()
    {
        if (strpos($this->order_params["form_url"], "/order/inline") == True)
            return "order";
        else
            return "create";
    }

    private function setUseTaxes()
    {
        $this->use_taxes = True;
    }

    private function getUseTaxes()
    {
        return $this->use_taxes;
    }

    private function setUseDelivery()
    {
        $this->use_delivery = True;
    }

    private function getUseDelivery()
    {
        return $this->use_delivery;
    }

    private function recalculateTaxes($item_pos)
    {
        //recalculate taxes
        switch($this->fiscal_cart[$item_pos]['tax']) {
            case "vat10":
                $this->fiscal_cart[$item_pos]['tax_sum'] = round((float)
                    (($this->fiscal_cart[$item_pos]['sum']/110)*10), 2);
                break;
            case "vat18":
                $this->fiscal_cart[$item_pos]['tax_sum'] = round((float)
                    (($this->fiscal_cart[$item_pos]['sum']/118)*18), 2);
                break;
        }
    }

    protected function setTaxes($sum, $tax_rate)
    {
        $this->showDebugInfo($tax_rate);
        $taxes = array("tax" => "none", "tax_sum" => 0);
        switch(number_format($tax_rate, 0, ".", "")) {
            case 10:
                $taxes["tax"] = "vat10";
                $taxes["tax_sum"] = round((float)(($sum/110)*10), 2);
                break;
            case 18:
                $taxes["tax"] = "vat18";
                $taxes["tax_sum"] = round((float)(($sum/118)*18), 2);
                break;
        }
        return $taxes;
    }

    protected function showDebugInfo($obj_to_debug)
    {
        echo "<pre>";
        var_dump($obj_to_debug);
        echo "</pre>";
    }
}
